/*
 * Decoder modifications to consider data loss while decoding frames. Source code compatible with JM 16.1
 * Author: Enrico Masala <masala@polito.it>, Politecnico di Torino, Jan 3, 2014
 *
 * Disclaimer:
 * This software is meant to be a prototype only, it is not optimized or fitted to any purposes.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 *
 * Citing this work:
 * We kindly ask people using this code in their research activity to cite a work
 * in which the approach employed by this software is used, e.g.,
 * E. Masala, A. Vesco, M. Baldi, J.C. De Martin,
 * "Optimized H.264 Video Encoding and Packetization for Video Transmission Over Pipeline Forwarding Networks",
 * IEEE Transactions on Multimedia, vol. 11, n. 5, Aug 2009, pp. 972-985.
 * 
 * This work has also been contributed to the VQEG JEG-Hybrid group, see also the meeting input document at:
 * ftp://vqeg.its.bldrdoc.gov/Documents/VQEG_Boulder_Jan14/MeetingFiles/
 * VQEG_JEG-Hybrid_2014__005_Politecnico_di_Torino_contribution_Simulation_of_Robust_H.264_AVC_Decoding.pdf
 * 
 * For more up-to-date information and publications please visit: http://media.polito.it/jeg
 *
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <limits.h>

#include <string.h>

#include "global.h"
// For the dec_picture variable
#include "image.h"
// For get_mem2Dpel
#include "memalloc.h"
// For getDpbSize()
#include "mbuffer.h"

#include "annexb.h"

#include "EM_JEG_modifications.h"


#define BITSTREAM_HEADER_SIZE 8
#define RTP_HEADER_SIZE 12

//#define MAX_MB_NUM ((11*4)*(9*4))
// FullHD  1920*1080 = 8160 MB
#define MAX_MB_NUM 8192
	
#define MAX_BUF (200+6*(MAX_MB_NUM))
char buffer[MAX_BUF];

static int isMBConcealed[MAX_MB_NUM];

EM_JEG_fd_t filedes;

static int current_slice_num = 0;
static EM_JEG_slice_info_t *current_slice_info;

int EM_JEG_debug_print = 0;

StorablePicture **EM_JEG_dpb;
EM_JEG_dpb_info_t *EM_JEG_dpb_info;
int *EM_JEG_debug_dpb_poc;
int EM_JEG_dpb_max;
int EM_JEG_dpb_size;
static int EM_JEG_frame_num = 0;
int EM_JEG_max_frame_num = 0;
int EM_JEG_flag_lossfirst = 0;
NALU_t *EM_JEG_nalu = NULL;
int EM_JEG_nalu_isLost = 0;

void EM_JEG_init(InputParameters *p_Inp) {
	filedes.loss = NULL;
	filedes.processed_bitstream = NULL;
	printf("EM_JEG: Info: If it happens, first frame is subject to losses? (0=no, 1=yes): %d\n", EM_JEG_flag_lossfirst);
        FILE *fptest = fopen(p_Inp->lossfile, "rt");
        if (fptest == NULL) {
                printf("EM_JEG: Error opening loss file '%s' from command line (-loss filename), trying loss.txt ...\n", p_Inp->lossfile);
		filedes.loss = EM_JEG_open_file_loss("loss.txt");
	} else {
		fclose(fptest);
		filedes.loss = EM_JEG_open_file_loss(p_Inp->lossfile);
	}

	if (filedes.loss) {
		EM_JEG_flag_lossfirst = p_Inp->lossfirst;
	}

	//printf("DEBUG EM_JEG p_Inp->processed_bitstream= '%s'\n", p_Inp->processed_bitstream);
	if (strlen(p_Inp->processed_bitstream)>0) {
        	filedes.processed_bitstream = EM_JEG_open_file_processed_bitstream(p_Inp->processed_bitstream);
		EM_JEG_nalu = AllocNALU(MAX_CODED_FRAME_SIZE); // It will be used to store a NALU after reading, before it is determined if it is lost or not
		//printf("EM_JEG_nalu->max_size=%d\n",EM_JEG_nalu->max_size);
		if ((EM_JEG_nalu->buf = (byte *)malloc (EM_JEG_nalu->max_size * sizeof(char))) == NULL) {
			printf("Cannot allocate the buffer for the EM_JEG_nalu element\n"); exit(1);
		}
		EM_JEG_nalu->len = 0;
	}

	int i;
	for (i=0; i<MAX_MB_NUM; i++) {
		isMBConcealed[i]=0;
	}

	current_slice_info = (EM_JEG_slice_info_t *)malloc(sizeof(EM_JEG_slice_info_t));
	if (!current_slice_info) {
		printf("Cannot allocate a EM_JEG_slice_info_t element\n"); exit(1);
	}
	current_slice_info->mb = (EM_JEG_mb_info_t *)malloc(sizeof(EM_JEG_mb_info_t)*MAX_MB_NUM);
	if (!current_slice_info->mb) {
		printf("Cannot allocate a EM_JEG_mb_info_t element\n"); exit(1);
	}
	current_slice_info->num_mb=0;

}

void EM_JEG_init_buffers(void) {
	// This is called ONLY when width and height etc. have been set

	//DecodedPicList
	EM_JEG_dpb_max = p_Dec->p_Vid->p_Dpb->size;
	//printf("DEBUG EM_JEG: EM_JEG_dpb_max=%d\n", EM_JEG_dpb_max);

	EM_JEG_dpb = (StorablePicture **)malloc(sizeof(StorablePicture *)*EM_JEG_dpb_max);
	if (!EM_JEG_dpb) { printf("ERROR: cannot allocate EM_JEG_dpb\n"); exit(1); }
	EM_JEG_dpb_info = (EM_JEG_dpb_info_t *)malloc(sizeof(EM_JEG_dpb_info_t)*EM_JEG_dpb_max);
	if (!EM_JEG_dpb_info) { printf("ERROR: cannot allocate EM_JEG_dpb_info\n"); exit(1); }
	EM_JEG_debug_dpb_poc = (int *)malloc(sizeof(int)*EM_JEG_dpb_max);
	if (!EM_JEG_debug_dpb_poc) { printf("ERROR: cannot allocate EM_JEG_debug_dpb_poc\n"); exit(1); }
	int i;
	int width = p_Dec->p_Vid->width;
	int height = p_Dec->p_Vid->height;
	int width_cr = p_Dec->p_Vid->width_cr;
	int height_cr = p_Dec->p_Vid->height_cr;

	for (i=0; i<EM_JEG_dpb_max; i++) {
		EM_JEG_dpb[i] = (StorablePicture *)malloc(sizeof(StorablePicture));
		if (!EM_JEG_dpb[i]) { printf("ERROR: cannot allocate EM_JEG_dpb[%d]\n",i); exit(1); }
		get_mem2Dpel(&(EM_JEG_dpb[i]->imgY), height, width);
		get_mem3Dpel(&(EM_JEG_dpb[i]->imgUV), 2, height_cr, width_cr);
		EM_JEG_dpb_info[i].poc=INT_MIN;
		EM_JEG_dpb_info[i].valid=0;
		EM_JEG_dpb_info[i].type=-1;
	}

}

void EM_JEG_print_BitstreamData(byte *buf, int len) {
	int i;
	int init=8;
	int end=8;
	if (len>init+end) {
		for (i=0; i<init; i++) {
			printf("%02x ",(int)(unsigned char)buf[i]);
		}
		printf("... ");
		for (i=len-end; i<len; i++) {
			printf("%02x ",(int)(unsigned char)buf[i]);
		}
	} else {
		for (i=0; i<len; i++) {
			printf("%02x ",(int)(unsigned char)buf[i]);
		}
	}
}

void EM_JEG_write_BitstreamData_if_needed(NALU_t *nalu) {
	//printf("DEBUG EM_JEG writing NALU with nalu->len = %d : %s \n", nalu->len, EM_JEG_nalu_isLost?"NO":"YES");
	if (! EM_JEG_nalu_isLost) {
		if (EM_JEG_nalu->len > 0 ) { // There is a previous NALU
			printf("Writing data: %5d ",nalu->len); EM_JEG_print_BitstreamData(nalu->buf, nalu->len); printf("\n");

			int ret = fwrite(nalu->buf, nalu->len, 1, filedes.processed_bitstream);
			if (ret!=1) {
				printf("ERROR: cannot write NALU in processed_bistream\n");
				exit(1);
			}
		}
	} else {
		EM_JEG_nalu_isLost = 0; // Reset the flag, previous NALU has not been written
	}
}

void EM_JEG_save_Annexb_NALU(byte *data, int len) {
	if (filedes.processed_bitstream) {
		EM_JEG_write_BitstreamData_if_needed(EM_JEG_nalu);

		// Now save current NALU, later it will be determined if this is considered lost or not
		if (len>0) {
			memcpy(EM_JEG_nalu->buf, data, len);
		}
		EM_JEG_nalu->len = len;
	}
}

void EM_JEG_save_RTP_Packet(byte *packetdata, int packetlen, int intime) {
	if (filedes.processed_bitstream) {
		EM_JEG_write_BitstreamData_if_needed(EM_JEG_nalu);

		// Now save current Packet Data, later it will be determined if this is considered lost or not
		if (packetlen>0) {
			// This is the RTP file format
			memcpy(EM_JEG_nalu->buf+0, &packetlen, 4);
			memcpy(EM_JEG_nalu->buf+4, &intime, 4);
			memcpy(EM_JEG_nalu->buf+8, packetdata, packetlen);
			EM_JEG_nalu->len = packetlen+8;
		} else {
			EM_JEG_nalu->len = 0;
		}
	}
}


void EM_JEG_free(void) {
	EM_JEG_close_file(filedes.loss);
	free(current_slice_info->mb);
	free(current_slice_info);
	if (filedes.processed_bitstream) {
		// Write the last NALU that has not been written
		EM_JEG_write_BitstreamData_if_needed(EM_JEG_nalu);

		EM_JEG_close_file(filedes.processed_bitstream);
		free(EM_JEG_nalu->buf);
		free(EM_JEG_nalu);
	}
}

void EM_JEG_dpb_print(void) {
	int i;
	for (i=0; i<EM_JEG_dpb_max; i++) {
		printf("(EM_JEG_dpb: %8p fr %3d valid %d type %d)\n",
			(void *)EM_JEG_dpb[i],
			EM_JEG_dpb_info[i].poc,
			EM_JEG_dpb_info[i].valid,
			EM_JEG_dpb_info[i].type);
	}
	printf("EM_JEG_dpb_size = %d\n", EM_JEG_dpb_size);
}


int int_cmp(const void *v1, const void *v2) { 
    return *(const int *)v1 - *(const int *)v2; 
} 

void EM_JEG_dpb_print_POCs(void) {
	int i;
	int idx=0;
	for (i=0; i<EM_JEG_dpb_max; i++) {
		if (EM_JEG_dpb_info[i].valid) {
			EM_JEG_debug_dpb_poc[idx] = EM_JEG_dpb_info[i].poc;
			idx++;
		}
	}

	qsort( EM_JEG_debug_dpb_poc, idx, sizeof(int), int_cmp);

	printf("EM_JEG_debug_dpb_poc: ");
	for (i=0; i<idx; i++) {
		printf("%d ", EM_JEG_debug_dpb_poc[i]);
	}
	printf("\n");
}


void EM_JEG_dpb_remove_oldest_picture(void) {
	int i;
	int poc_old = INT_MAX;
	int idx= -1;
	for (i=0; i<EM_JEG_dpb_max; i++) {
		if (EM_JEG_dpb_info[i].valid>0) {
			if (EM_JEG_dpb_info[i].poc < poc_old) {
				poc_old = EM_JEG_dpb_info[i].poc;
				idx=i;
			}
		}
	}
	if (idx<0) {
		printf("ERROR: EM_JEG_dpb: cannot remove oldest picture: cannot find it\n"); exit(1);
	}
	EM_JEG_dpb_info[idx].poc=INT_MIN;
	EM_JEG_dpb_info[idx].valid=0;
	EM_JEG_dpb_info[idx].type=-1;
	EM_JEG_dpb_size--;
}


int EM_JEG_dpb_find_free_element(void) {
	int idx=-1;
	int i;
	for (i=0; i<EM_JEG_dpb_max; i++) {
		if (EM_JEG_dpb_info[i].valid==0) {
			return i;
		}
	}
	return idx;
}

int EM_JEG_dpb_min_POC(void) {
	int poc_min=INT_MAX;
	//int idx=-1;
	int i;
	for (i=0; i<EM_JEG_dpb_max; i++) {
		if (EM_JEG_dpb_info[i].valid>0) {
			if (EM_JEG_dpb_info[i].poc < poc_min) {
				poc_min = EM_JEG_dpb_info[i].poc;
				//idx=i;
			}
		}
	}
	return poc_min;
}


// Find the frame with the highest poc < poc and type == B (isB=1) or IoP (isB=0)
// This is useful to conceal the frame
// NB: If poc < EM_JEG_dpb_min_POC() there will be no solution, so change poc to INT_MAX so that is it like a new poc greater than all is to be concealed
int EM_JEG_dpb_find_maxframe_and_type(int poc, int isIorP) {
	int poc_max=INT_MIN;
	int idx=INT_MIN;
	int i;
	//printf("DEBUG EM_JEG: EM_JEG_dpb_find_maxframe_and_type( poc= %d EM_JEG_dpb_min_POC()=%d )\n", poc, EM_JEG_dpb_min_POC() );
	if (poc < EM_JEG_dpb_min_POC()) {
		poc = INT_MAX;
	}
	for (i=0; i<EM_JEG_dpb_max; i++) {
		if (EM_JEG_dpb_info[i].valid>0) {
			if (EM_JEG_dpb_info[i].poc < poc) {
				if ( !isIorP || (isIorP &&  (EM_JEG_dpb_info[i].type == I_SLICE || EM_JEG_dpb_info[i].type == P_SLICE)) ) {
					if (EM_JEG_dpb_info[i].poc > poc_max) {
						poc_max = EM_JEG_dpb_info[i].poc;
						idx=i;
					}
				}
			}
		}
	}
	//if (idx!=INT_MIN) { printf("DEBUG EM_JEG: EM_JEG_dpb_find_maxframe_and_type(poc=%d, isIorP=%d) = %d (poc=%d)\n", poc, isIorP, idx, EM_JEG_dpb_info[idx].poc); } else { printf("DEBUG EM_JEG: EM_JEG_dpb_find_maxframe_and_type(poc=%d, isIorP=%d) = %d (poc= -)\n", poc, isIorP, idx); }
	return idx;
}


// Change the poc of all valid frames in EM_JEG_dpb of the specified amount
void EM_JEG_dpb_shift_all_POC(int shift_val) {
	int i;
	for (i=0; i<EM_JEG_dpb_max; i++) {
		if (EM_JEG_dpb_info[i].valid>0) {
			EM_JEG_dpb_info[i].poc += shift_val;
		}
	}
}



int EM_JEG_isMBConcealed(int idx) {
	return isMBConcealed[idx];
}


void EM_JEG_apply_slice_concealment(Slice *currSlice) {
	int i;
	int loss_event;

	VideoParameters *img = currSlice->p_Vid;
 
	loss_event = EM_JEG_get_loss_event();

	//if (loss_event) {
	//	printf("DEBUG EM_JEG_apply_slice_concealment: img->type=%d\n",img->type);
	//	printf("Slice content: ");
	//	for (i=0; i<current_slice_info->num_mb; i++) {
	//		printf("%d ", current_slice_info->mb[i].mb_addr);
	//	}
	//	printf("\n");
	//}

	//int potential_poc_num__to_be_concealed = img->dec_picture->frame_poc;
	//printf("EM_JEG DEBUG: potential_poc_num__to_be_concealed = %d\n", potential_poc_num__to_be_concealed);

	if (loss_event) {
		EM_JEG_nalu_isLost = 1;  // Do not write the NALU in the processed_bitstream

		StorablePicture* refPic;

		//seq_parameter_set_rbsp_t *active_sps = img->active_sps;
		// for POC mode 0:
		//unsigned int MaxPicOrderCntLsb = (1<<(active_sps->log2_max_pic_order_cnt_lsb_minus4+4));
		//printf("DEBUG MaxPicOrderCntLsb=%d\n", MaxPicOrderCntLsb);


		//listX[0] list0 for frame
		//listX[1] list1 for frame
		//listX[2] list0 for current_field==top
		//listX[3] list1 for current_field==top
		//listX[4] list0 for current_field==bottom
		//listX[5] list1 for current_field==bottom

		//printf("DEBUG EM_JEG: listXsize[0]=%d listXsize[1]=%d\n", listXsize[0], listXsize[1]);
         
		int poc_num__to_be_concealed = img->dec_picture->frame_poc;
		//printf("EM_JEG DEBUG: poc_num__to_be_concealed = %d\n", poc_num__to_be_concealed);

		//if (EM_JEG_debug_print) printf("poc_num__to_be_concealed = %d\n", poc_num__to_be_concealed);

		// It should work also when POC wraps around, due to EM_JEG_dpb_shift_all_POC() (FIXME: which value to use? ~1000, i.e., 500 frames?)
		int idx = INT_MIN;
		if ( img->type==I_SLICE || img->type==P_SLICE ) {
			idx = EM_JEG_dpb_find_maxframe_and_type(poc_num__to_be_concealed, 0);
		} else {
			// This is a B_SLICE

			// OLD COMMENT
			// list1 do NOT contain the last decoded frame. For this, you need to look into the DPB.
			// Only list0 is used for usual buffer management as in MPEG-2 even for backward-predicted frames.
			// list0 and list1 are there for hierarchical frame coding etc.
			// But DPB management is fairly complex (short term references, long term refeences, etc...)
			// Then, after a frame is added with  store_picture_in_dpb()  the buffer is reorganized to make room for new frames
			// by outputting the one not used for reference (in the correct order in case P and B are present...) and then added.
			// Therefore it is very difficult do determine the index at which the frame is stored into the DPB (and it may change at next step).
			// For this reason, a copy is kept locally to simplify the code.
			// Function  EM_JEG_save_ref_frame() implement this

			// NEW COMMENT
			// a DPB is needed anyway to ensure correct operation in any conditions (e.g., IIIII etc.)

			// Use this code to use the last reference frame, not the last frame as concealment
			//idx = EM_JEG_dpb_find_maxframe_and_type(poc_num__to_be_concealed, 1);

			idx = EM_JEG_dpb_find_maxframe_and_type(poc_num__to_be_concealed, 0);
		}

                if (EM_JEG_dpb_size > 0) {
			refPic = EM_JEG_dpb[idx];
		} else {
			refPic = NULL;  // No pictures decoded yet (to use for concealment)
		}

		EM_JEG_debug_print=1;
		for (i=0; i<current_slice_info->num_mb; i++) {
			// dec_picture is a global variable (declared in image.h)

			EM_JEG_apply_mb_concealment(img, img->dec_picture->imgY, img->dec_picture->imgUV, refPic, current_slice_info->mb[i].mb_addr);
			EM_JEG_debug_print=0;
		}
	}

	// Reset the content of the current slice: prepare for the next one
	current_slice_info->num_mb=0;

	// increment total number of slices
	current_slice_num++;
}


char *EM_JEG_toString_imgtype(int type) {
	if (type==0)
		return "P ";
	else if (type==1)
		return "B ";
	else if (type==2)
		return "I ";
	else if (type==3)
		return "SP";
	else if (type==4)
		return "SI";
	else
		return "? ";
}

void EM_JEG_compute_mb_coordinates(VideoParameters *img, int *mb_x, int *mb_y, int mb_address) {
	int mb_width;
	//int mb_height;
	mb_width = img->PicWidthInMbs;
	//mb_height = img->PicHeightInMbs;
	*mb_y = mb_address / mb_width;
	*mb_x = mb_address - *mb_y * mb_width;
	return;
}

void EM_JEG_apply_mb_concealment(VideoParameters *img, imgpel **imgY, imgpel ***imgUV, StorablePicture *refPic, int mb_address) {
	int mb_x;
	int mb_y;
	int mb_x_pel;
	int mb_y_pel;
	int i,j;
	
	////if (EM_JEG_debug_print) printf("concealing with POC = %d %d\n", refPic->poc, EM_JEG_dpb_info[idx].poc);
	//if (EM_JEG_debug_print) printf("DEBUG EM_JEG: concealing with POC = %d\n", refPic->poc);
	
	EM_JEG_compute_mb_coordinates(img, &mb_x, &mb_y, mb_address);
	//printf("EM_JEG: mb_x=%d,mb_y=%d\n",mb_x,mb_y);
	mb_x_pel = mb_x * 16;
	mb_y_pel = mb_y * 16;

	//if (poc_num__to_be_concealed != 0) 
	if (refPic!=NULL) {
		isMBConcealed[mb_address]=1;
		for (j=0; j<16; j++) {
			for (i=0; i<16; i++) {
				imgY[mb_y_pel+j][mb_x_pel+i] = refPic->imgY[mb_y_pel+j][mb_x_pel+i];
			}
		}
		for (j=0; j<8; j++) {
			for (i=0; i<8; i++) {
				imgUV[0][mb_y_pel/2+j][mb_x_pel/2+i] = refPic->imgUV[0][mb_y_pel/2+j][mb_x_pel/2+i];
				imgUV[1][mb_y_pel/2+j][mb_x_pel/2+i] = refPic->imgUV[1][mb_y_pel/2+j][mb_x_pel/2+i];
			}
		}
	} else {
		if (EM_JEG_flag_lossfirst) {  // EM_JEG: do only if something in the first frame has to be lost
			// In case of first frame
			// Assuming concealment with value 128
			isMBConcealed[mb_address]=1;
			for (j=0; j<16; j++) {
				for (i=0; i<16; i++) {
					imgY[mb_y_pel+j][mb_x_pel+i] = 128;
				}
			}
			for (j=0; j<8; j++) {
				for (i=0; i<8; i++) {
					imgUV[0][mb_y_pel/2+j][mb_x_pel/2+i] = 128;
					imgUV[1][mb_y_pel/2+j][mb_x_pel/2+i] = 128;
				}
			}
		}
	}

}

void copy_frame(imgpel **dst, imgpel **src, int w, int h) {
	int i,j;

	for (j=0;j<h;j++) {
		for (i=0;i<w;i++) {
			dst[j][i]=src[j][i];
		}
	}
}

//swap(void *a, void *b) int t = *a; *a=*b; *b=t;
void EM_JEG_swap_sp(StorablePicture **p1, StorablePicture **p2) {
	StorablePicture *tmp = *p1;
	*p1 = *p2;
	*p2 = tmp;
}
	
void EM_JEG_init_frame(StorablePicture *dec_picture) {
	//int potential_poc_num__to_be_concealed = dec_picture->frame_poc;
	//printf("EM_JEG DEBUG: init_frame potential_poc_num__to_be_concealed = %d\n", potential_poc_num__to_be_concealed);

	// Check if the new frame has POC < than the older frames (e.g., a new IDR with poc=0 has arrived)
	int EM_JEG_min_POC_in_dpb = EM_JEG_dpb_min_POC();
	if (EM_JEG_dpb_size > 0) {
		if (dec_picture->poc <= EM_JEG_min_POC_in_dpb) {
			// Shift all poc values in the buffer so that the new is the highest (e.g., a new IDR with poc=0 has arrived)
			EM_JEG_dpb_shift_all_POC(-500); // Must be much lower than 0 (IDR has 0, but some B frames not yet read can be negative)
		}
	}
}

void EM_JEG_save_frame(StorablePicture *dec_picture) {
	//printf("DEBUG EM_JEG: save_frame() called\n");
	//printf("DEBUG EM_JEG: EM_JEG_dpb_size: %d\n", EM_JEG_dpb_size);

	//EM_JEG_dpb_print();
	//EM_JEG_dpb_print_POCs();

	if (EM_JEG_dpb_size==EM_JEG_dpb_max) {
		//printf("DEBUG EM_JEG: removing oldest picture\n");
		EM_JEG_dpb_remove_oldest_picture();
	}
	
	int idx = EM_JEG_dpb_find_free_element();
	if (idx<0) {
		printf("ERROR: cannot store picture in EM_JEG_dpb: buffer is full\n"); exit(1);
	}


	// add picture to EM_JEG_dpb in position idx
	StorablePicture *p = EM_JEG_dpb[idx];

	int width = p_Dec->p_Vid->width;
	int height = p_Dec->p_Vid->height;
	int width_cr = p_Dec->p_Vid->width_cr;
	int height_cr = p_Dec->p_Vid->height_cr;

	copy_frame(p->imgY,dec_picture->imgY, width, height);
	copy_frame(p->imgUV[0],dec_picture->imgUV[0], width_cr, height_cr);
	copy_frame(p->imgUV[1],dec_picture->imgUV[1], width_cr, height_cr);
	p->poc = dec_picture->poc;
	//printf("DEBUG EM_JEG: adding picture to my dpb: poc= %d\n", p->poc);
	// Remember to copy also other fields if needed... (e.g., MV for concealment)

	EM_JEG_dpb_info[idx].poc = dec_picture->poc;
	EM_JEG_dpb_info[idx].valid = 1;
	EM_JEG_dpb_info[idx].type = dec_picture->slice_type;
	EM_JEG_dpb_size++;

	EM_JEG_frame_num++;

	// Reset MB concealment flags
	int i;
	for (i=0; i<MAX_MB_NUM; i++) {
		isMBConcealed[i]=0;
	}

	//EM_JEG_dpb_print();
	//EM_JEG_dpb_print_POCs();
}



void EM_JEG_add_new_mb(int mb_addr) {
	if (current_slice_info->num_mb < MAX_MB_NUM) {
		current_slice_info->mb[current_slice_info->num_mb].mb_addr = mb_addr;  // May not be correct if FMO is enabled
		current_slice_info->mb[current_slice_info->num_mb].type = -1; // Currently unused
		current_slice_info->num_mb++;
	} else {
		printf("Cannot add macroblock address to current_slice_info: increase MAX_MB_NUM\n"); exit(1);
	}
	return;
}


#define MAX_BUF_EVENT 1024

int EM_JEG_get_loss_event(void) {
	int val;
	char buf[MAX_BUF_EVENT];

	if (fgets(buf, MAX_BUF_EVENT, filedes.loss) == NULL) {
		printf("EM_JEG: Error: Premature end on file of loss events!\n");
		exit(1);
	}
	val = buf[0] - '0';

	if (val != 0 && val != 1) {
		printf("EM_JEG: Error: event must be 0 or 1! (Read: 0x%02x = '%c' = %3d)\n", val, (char)val, val);
		exit(1);
	}

	return val;
}




// #######################################################
// ###  Utility functions
// #######################################################

FILE *EM_JEG_open_file_processed_bitstream(char *fname) {

	FILE *fp = fopen(fname, "wb");	
	if (fp == NULL) {
		printf("EM_JEG: Error opening processed bitstream '%s' for writing\n", fname);
		exit(1);
	}
	printf("EM_JEG: File %s opened!\n", fname);
	return fp;
}


FILE *EM_JEG_open_file_loss(char *fname) {

	FILE *fp = fopen(fname, "rt");	
	if (fp == NULL) {
		printf("EM_JEG: Error opening loss file '%s'\n", fname);
		exit(1);
	}
	printf("EM_JEG: File %s opened!\n", fname);
	return fp;
}

void EM_JEG_close_file(FILE *fp) {
	if (fp) {
		fclose(fp);
	}
}

